/*
   (C) Andreas F. Borchert, 2002, <unrarlib@andreas-borchert.de>.

   unrar is free software; you can redistribute it and/or modify it under
   the terms of the GNU Library General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   unrar is distributed in the hope that it will be useful, but WITHOUT
   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
   or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Library General
   Public License for more details.

   You should have received a copy of the GNU Library General Public
   License along with this library; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/*

This is manual page is in POD format:
   http://www.perldoc.com/perl5.6.1/pod/perlpod.html
Use pod2<something> to convert it to your favourite
output format.

=head1 NAME

unrar -- list rar archives and extract files from them

=head1 SYNOPSIS

B<unrar> [B<-c>] [B<-l>] [B<-p> I<password>] [B<-v>] [B<-x>] I<archive>

=head1 DESCRIPTION

B<unrar> tries it luck in listing or unpacking I<archive> which
is assumed to be in RAR 2.0 format. Following options are
supported:

=over 4

=item B<-c>

Permit extracted files to overwrite existing files. By default,
existing files are not touched.

=item B<-l>

Request a long listing that gives the number of bytes for each
entry within the archive.

=item B<-p> I<password>

Specify a password. Extraction fails if CRC checks are enabled
in the library and if the password does not match. By default,
the password is the empty string.

=item B<-v>

Ask for more verbosity. This adds the file mode to the long
listing and some activity reports if in extraction mode.

=item B<-x>

Extract all directories and files from the archive. By default,
just the entries are listed.

=back

=head1 DIAGNOSTICS

If successful, B<unrar> exits with 0. In case of usage errors, an exit
code of 111 is returned. Other errors lead to an exit code of 1. Error
messages are printed to stderr. Unfortunately, B<unrar> is not able to
print detailed error messages in case of problems with the RAR format
as the library provides no support for this.

=head1 AUTHOR

This utility was written by Andreas Borchert and is based upon unrarlib,
see http://www.unrarlib.org/. My contact:
unrarlib@andreas-borchert.de

=head1 DISCLAIMER

This is a hack that was developed to unpack a RAR file on a system
where no ready-to-use utilities for this task existed.

Please note that unrarlib has severe security problems, i.e. there are
several buffers whose bounds are not verified. Malicious RAR files can
be crafted that allow to execute arbitrary code. You have been warned!

=head1 BUGS

See disclaimers.

I do not know how to distinguish directories from regular files
in the archive reliably. It worked for my samples. My assumptions
might fail in your RAR archive.

This utility assumes that the RAR archive was generated on a
system where backslashes are used as directory delimiter and
converts them to slashes under the assumption that we live on
a UNIX system.

=cut

*/

#include <errno.h>
#include <fcntl.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unrarlib.h>
#include <unistd.h>

#include <sys/types.h>
#include <sys/stat.h>

#define USAGE_ERROR 111
#define FATAL_ERROR 1
#define OK 0

char usage[] = "Usage: %s [-c] [-l] [-p password] [-v] [-x] archive\n";
char* cmdname; /* basename of argv[0] */
int clobber = 0;      /* -c */
int long_listing = 0; /* -l */
char* password = 0;   /* -p password */
int verbose = 0;      /* -v */
int extract = 0;      /* -x */

int list_archive(const char* filename) {
   int number_of_files;
   ArchiveList_struct* list;

   errno = 0; /* check for system call errors */
   number_of_files = urarlib_list(filename, &list);
   if (number_of_files <= 0) {
      fprintf(stderr, "%s: unable to open %s as RAR file\n",
	 cmdname, filename);
      if (errno) {
	 perror(filename);
      }
      return FATAL_ERROR;
   }

   while (list != 0) {
      if (long_listing) {
	 if (verbose) {
	    printf("%5lu ",  list->item.FileAttr);
	 }
	 printf("%10lu ", list->item.UnpSize);
      }
      puts(list->item.Name);
      list = list->next;
   }
   return OK;
}

int extract_archive(const char* filename, const char* password) {
   int number_of_files;
   ArchiveList_struct* list;

   errno = 0; /* check for system call errors */
   number_of_files = urarlib_list(filename, &list);
   if (number_of_files <= 0) {
      fprintf(stderr, "%s: unable to open %s as RAR file\n",
	 cmdname, filename);
      if (errno) {
	 perror(filename);
      }
      return FATAL_ERROR;
   }

   while (list != 0) {
      void* data;
      unsigned long data_size;
      char* path;
      char* cp;

      if (verbose) {
	 if (long_listing) {
	    printf("%10lu ", list->item.UnpSize);
	 }
	 fputs(list->item.Name, stdout);
      }

      /*
       * copy list->item.Name to path and fix directory delimiters
       *
       * This could be done better if all my questions about
       * RAR would be answered. This hack simply converts
       * backslashes (assuming MS-DOS/Windows as the originating
       * system) to slashes (assuming UNIX).
       */
      path = strdup(list->item.Name);
      if (path == 0) {
	 fprintf(stderr, "%s: unable to alloc memory\n", cmdname);
	 return FATAL_ERROR;
      }
      for (cp = path; *cp; ++cp) {
	 if (*cp == '\\') {
	    *cp = '/';
	 }
      }

      /*
       * I have no idea about the file attributes and was
       * unable to find documentation for it.
       * From some samples I have the guess that in case
       * of directories the 0x10 bit is set.
       */
      if (list->item.FileAttr & 0x10) {
	 errno = 0;
	 if (mkdir(path, 0777) < 0) {
	    if (errno == EEXIST) {
	       if (verbose) {
		  printf(" ... exists already\n");
	       }
	    } else {
	       if (verbose) {
		  printf(" ... cannot be created\n");
	       }
	       fprintf(stderr, "%s: %s: cannot be created\n",
		  cmdname, path);
	       perror(path);
	       return FATAL_ERROR;
	    }
	 } else {
	    if (verbose) {
	       printf(" ... created\n");
	    }
	 }
      } else {
	 /* looks like a regular file, let us hope for the best... */
	 if (urarlib_get(&data, &data_size, list->item.Name,
	       filename, password)) {
	    ssize_t nbytes = 0;
	    int fcntl_flags = O_WRONLY | O_CREAT;
	    if (clobber) {
	       fcntl_flags |= O_TRUNC;
	    } else {
	       fcntl_flags |= O_EXCL;
	    }
	    int fd = open(path, fcntl_flags, 0666);
	    if (fd < 0) {
	       if (verbose) {
		  printf(" ... cannot be created\n");
	       }
	       fprintf(stderr, "%s: %s: cannot be created\n",
		  cmdname, path);
	       perror(path);
	       return FATAL_ERROR;
	    }
	    while (nbytes < data_size) {
	       int written = write(fd, data + nbytes, data_size - nbytes);
	       if (written < 0) {
		  if (verbose) {
		     printf(" ... write error\n");
		  }
		  fprintf(stderr, "%s: %s: write error\n",
		     cmdname, path);
		  perror(path);
		  return FATAL_ERROR;
	       }
	       nbytes += written;
	    }
	    if (close(fd) < 0) {
	       if (verbose) {
		  printf(" ... write error\n");
	       }
	       fprintf(stderr, "%s: %s: write error\n",
		  cmdname, path);
	       perror(path);
	       return FATAL_ERROR;
	    }
	    printf(" ... extracted\n");
	 } else {
	    if (verbose) {
	       printf(" ... failed\n");
	    }
	 }
      }

      list = list->next;
   }
   return OK;
}

int main(int argc, char** argv) {
   int flag;

   /* extract basename out of argv[0] */
   cmdname = strrchr(argv[0], '/');
   if (cmdname == 0) {
      cmdname = argv[0];
   } else {
      ++cmdname;
   }

   opterr = 0; /* disable error messages by getopt */
   while ((flag = getopt(argc, argv, "clp:vx")) != EOF) {
      switch(flag) {
	 case 'c':
	    clobber = 1; break;
	 case 'l':
	    long_listing = 1; break;
	 case 'p':
	    password = optarg; break;
	 case 'v':
	    verbose = 1; break;
	 case 'x':
	    extract = 1; break;
	 case '?':
	    fprintf(stderr, usage, cmdname); exit(USAGE_ERROR);
      }
   }

   /* just the archive name has to be left on the command line */
   if (optind != argc - 1) {
      fprintf(stderr, usage, cmdname); exit(USAGE_ERROR);
   }

   if (extract) {
      return extract_archive(argv[optind], password);
   } else {
      return list_archive(argv[optind]);
   }
}

